#include <Python.h>
#include <sys/mman.h>
#include "lpyhook.h"

#define CATLIZED_SIGN "__catlized"
#define CAPI_METHOD "exc_capi"

extern PyObject* _PyFunction_FastCallKeywords(PyObject *func, PyObject *const *stack, Py_ssize_t nargs, PyObject *kwnames);

PyObject *
hookify_PyFunction_FastCallKeywords(PyObject *func, PyObject *const *stack, Py_ssize_t nargs, PyObject *kwnames)
{
    __asm__("NOP");
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *kwdefs, *closure, *name, *qualname;
    PyObject **d;
    Py_ssize_t nkwargs = (kwnames == NULL) ? 0 : PyTuple_GET_SIZE(kwnames);
    Py_ssize_t nd;
    PyObject *w, *args, *instance, *catlizor, *meth_name, *tracked;
    int i, n, pre = 0, on_call = 0, post = 0;
    assert(PyFunction_Check(func));
    assert(nargs >= 0);
    assert(kwnames == NULL || PyTuple_CheckExact(kwnames));
    assert((nargs == 0 && nkwargs == 0) || stack != NULL);

    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
    closure = PyFunction_GET_CLOSURE(func);
    name = ((PyFunctionObject *)func) -> func_name;
    qualname = ((PyFunctionObject *)func) -> func_qualname;

    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = PyTuple_GET_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
    
    if (nargs > co->co_argcount) {
        n = co->co_argcount;
    }
    else {
        n = nargs;
    }
    
    args = PyList_New(n);
    for (i = 0; i < n; i++) {
        w = stack[i];
        Py_INCREF(w);
        PyList_SetItem(args, i, w);
    }
    
    instance = PyList_GetItem(args, 0);
    if (PyObject_HasAttrString(instance, CATLIZED_SIGN)){
        catlizor = PyObject_GetAttrString(instance, CATLIZED_SIGN);
        meth_name = PyObject_GetAttrString(func, "__name__");
        tracked = PyObject_CallMethod(catlizor, "tracked", "(OO)", meth_name, Py_True);
        if (PySet_Check(tracked) && PySet_Size(tracked) > 0){
            pre = PySet_Contains(tracked, PyLong_FromLong(0));
            on_call = PySet_Contains(tracked, PyLong_FromLong(1));
            post = PySet_Contains(tracked, PyLong_FromLong(2));
        }

    }
    
    if (pre)
        PyObject_CallMethod(catlizor, CAPI_METHOD, "(iOOO)", 0, func, args, Py_None);

    PyObject *result = _PyEval_EvalCodeWithName((PyObject*)co, globals, (PyObject *)NULL,
                                    stack, nargs,
                                    nkwargs ? &PyTuple_GET_ITEM(kwnames, 0) : NULL,
                                    stack + nargs,
                                    nkwargs, 1,
                                    d, (int)nd, kwdefs,
                                    closure, name, qualname);

    if (on_call)
        PyObject_CallMethod(catlizor, CAPI_METHOD, "(iOOO)", 1, func, args, result);
    
    if (post)
        PyObject_CallMethod(catlizor, CAPI_METHOD, "(iOOO)", 2, func, args, Py_None);
    
    return result;
}

static PyMethodDef module_methods[] = {
    {NULL}
};


static struct PyModuleDef hookify =
{
    PyModuleDef_HEAD_INIT,
    "hookify",
    NULL,
    -1, 
    module_methods
};

PyMODINIT_FUNC PyInit_hookify(void) {
    __asm__("");
    lpyhook(_PyFunction_FastCallKeywords, &hookify_PyFunction_FastCallKeywords);
    return PyModule_Create(&hookify);
}

